/*
  nntpd -- the NNTP server
  See AUTHORS for copyright holders and contributors.
  See README for restrictions on the use of this software.
*/

#include "leafnode.h"

#include <unistd.h>
#include <assert.h>

#include "critmem.h"
#include "ln_log.h"
#include "get.h"
#include "attributes.h"
#include "mastring.h"
#include "h_error.h"
#include "masock.h"

/* FIXME: write wrapper for this fellow */
#ifdef SOCKS
#include <socks.h>
#endif

#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <netdb.h>
#include <setjmp.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <time.h>
#include <sys/time.h>
#include <utime.h>

#ifdef DEBUG_DMALLOC
#include <dmalloc.h>
#endif

#define MAXLINELENGTH 1000

/* for authorization: */
#define P_NOT_SUPPORTED	503
#define P_SYNTAX_ERROR	501
#define P_REJECTED	482
#define P_CONTINUE	381
#define P_ACCEPTED	281

static char *generateMessageID(void);
static FILE *fopenart(const struct newsgroup *group, const char *,
		      unsigned long *);
static FILE *buildpseudoart(const char *grp);
static FILE *fopenpseudoart(const struct newsgroup *group, const char *arg,
			    const unsigned long article_num);
static void list(struct newsgroup *ng, int what, char *pattern);
static void main_loop(void);	/* main program loop */
static void doarticle(const struct newsgroup *group, const char *arg, int what,
		      unsigned long *);
static struct newsgroup *dogroup(const char *, unsigned long *);
static void dohelp(void);
static void domode(const char *arg);
static void domove(const struct newsgroup *group, int, unsigned long *);
static void dolist(char *p);
static void donewgroups(const char *);
static void donewnews(char *);
static void dopost(void);
static void doxhdr(const struct newsgroup *group, const char *,
		   unsigned long artno);
static void doxpat(const struct newsgroup *group, const char *,
		   unsigned long artno);
static void doxover(const struct newsgroup *group, const char *, unsigned long);
static void doselectedheader(const struct newsgroup *, const char *,
			     const char *, struct stringlist *,
			     unsigned long *);
static struct newsgroup *dolistgroup(struct newsgroup *group, const char *,
				     unsigned long *);
static int markinterest(const struct newsgroup *);
static int allowposting(void);
static int isauthorized(void);
static void doauthinfo(const char *arg);
static int dorange(const char *arg,
		   unsigned long *a, unsigned long *b,
		   unsigned long artno, unsigned long lo, unsigned long hi);

static const struct newsgroup *xovergroup;	/* FIXME */
static struct stringlist *users = NULL;	/* FIXME */
				/* users allowed to use the server */
int debug = 0;
static int authflag = 0;	/* TRUE if authenticated */
static time_t gmt_off;		/* offset between localtime and GMT in sec */

/*
 * this function avoids the continuous calls to both ln_log and printf
 * it also appends \r\n automagically
 * slightly modified from a function nntp_log_and_reply() written by
 * Matthias Andree <ma@dt.e-technik.uni-dortmund.de> (c) 1999
 */
static void nntpprintf(const char *fmt, ...)
    __attribute__ ((format(printf, 1, 2)));
static void
nntpprintf(const char *fmt, ...)
{
    char buffer[1024];
    va_list args;

    va_start(args, fmt);
    vsnprintf(buffer, sizeof(buffer), fmt, args);
    if (debugmode & DEBUG_IO)
	ln_log(LNLOG_SDEBUG, LNLOG_CALL, ">%s", buffer);
    printf("%s\r\n", buffer);
    fflush(stdout);
    va_end(args);
}

static void
main_loop(void)
{
    /* "global" state */
    static unsigned long artno = 0;
    static struct newsgroup *group = 0;

    /* locals */
    char *arg, *cmd;
    int n;

    while (fflush(stdout), (cmd = mgetaline(stdin))) {
	/* collect possible returned children */
	while (waitpid(-1, 0, WNOHANG) > 0);

	if (debugmode & DEBUG_NNTP && !(debugmode & DEBUG_IO))
	    ln_log(LNLOG_SDEBUG, LNLOG_CTOP, "< %s", cmd);

	n = strlen(cmd);
	if (n == 0)
	    continue;		/* necessary for netscape to be quiet */
	else if (n > MAXLINELENGTH) {
	    /* ignore attempts at buffer overflow */
	    nntpprintf("500 Dazed and confused");
	    fflush(stdout);
	    continue;
	}
	cmd = critstrdup(cmd, "main_loop");
	/* parse command line */
	n = 0;
	while (isspace((unsigned char)cmd[n]))
	    n++;
	while (isalpha((unsigned char)cmd[n]))
	    n++;
	while (isspace((unsigned char)cmd[n]))
	    cmd[n++] = '\0';
	arg = cmd + n;
	while (cmd[n])
	    n++;
	n--;
	while (isspace((unsigned char)cmd[n]))
	    cmd[n--] = '\0';
	if (!strcasecmp(cmd, "quit")) {
	    nntpprintf("205 Always happy to serve!");
	    free(cmd);
	    return;
	}
	if (!strcasecmp(cmd, "article")) {
	    if (isauthorized())
		doarticle(group, arg, 3, &artno);
	} else if (!strcasecmp(cmd, "head")) {
	    if (isauthorized())
		doarticle(group, arg, 2, &artno);
	} else if (!strcasecmp(cmd, "body")) {
	    if (isauthorized())
		doarticle(group, arg, 1, &artno);
	} else if (!strcasecmp(cmd, "stat")) {
	    if (isauthorized())
		doarticle(group, arg, 0, &artno);
	} else if (!strcasecmp(cmd, "help")) {
	    dohelp();
	} else if (!strcasecmp(cmd, "ihave")) {
	    nntpprintf("500 IHAVE is for big news servers");
	} else if (!strcasecmp(cmd, "last")) {
	    if (isauthorized())
		domove(group, -1, &artno);
	} else if (!strcasecmp(cmd, "next")) {
	    if (isauthorized())
		domove(group, 1, &artno);
	} else if (!strcasecmp(cmd, "list")) {
	    if (isauthorized())
		dolist(arg);
	} else if (!strcasecmp(cmd, "mode")) {
	    if (isauthorized())
		domode(arg);
	} else if (!strcasecmp(cmd, "newgroups")) {
	    if (isauthorized())
		donewgroups(arg);
	} else if (!strcasecmp(cmd, "newnews")) {
	    if (isauthorized())
		donewnews(arg);
	} else if (!strcasecmp(cmd, "post") && allowposting()) {
	    if (isauthorized())
		dopost();
	} else if (!strcasecmp(cmd, "slave")) {
	    if (isauthorized()) {
		nntpprintf("202 Cool - I always wanted a slave");
	    }
	} else if (!strcasecmp(cmd, "xhdr")) {
	    if (isauthorized())
		doxhdr(group, arg, artno);
	} else if (!strcasecmp(cmd, "xpat") || !strcasecmp(cmd, "pat")) {
	    if (isauthorized())
		doxpat(group, arg, artno);
	} else if (!strcasecmp(cmd, "xover") || !strcasecmp(cmd, "over")) {
	    if (isauthorized())
		doxover(group, arg, artno);
	} else if (!strcasecmp(cmd, "listgroup")) {
	    if (isauthorized())
		group = dolistgroup(group, arg, &artno);
	} else if (!strcasecmp(cmd, "group")) {
	    if (isauthorized())
		group = dogroup(arg, &artno);
	} else if (!strcasecmp(cmd, "authinfo")) {
	    doauthinfo(arg);
	} else {
	    nntpprintf("500 Unknown command");
	}
	free(cmd);
    }
    ln_log(LNLOG_SWARNING, LNLOG_CTOP,
	   "Warning: premature EOF in client input.");
#ifndef NOT_YET
    /* this will become obsolete when the new NNTP standards are
       published around December 2001 or Early 2002, we'd drop the 400
       error message then. Until then, we send it in compliance with
       RFC-977 p. 23 */
    nntpprintf("400 service discontinued: EOF");
#endif
}

/*
 * pseudo article stuff
 */
static void
fprintpseudobody(FILE * pseudoart, const char *groupname)
{
    FILE *f;
    char *l, *cl, *c;

    if (pseudofile && ((f = fopen(pseudofile, "r")) != NULL)) {
	debug = 0;
	while ((l = getaline(f)) != NULL) {
	    cl = l;
	    while ((c = strchr(cl, '%')) != NULL) {
		if (strncmp(c, "%%", 2) == 0) {
		    *c = '\0';
		    fprintf(pseudoart, "%s%%", cl);
		    cl = c + 2;
		} else if (strncmp(c, "%server", 7) == 0) {
		    *c = '\0';
		    fprintf(pseudoart, "%s%s", cl, owndn ? owndn : fqdn);
		    cl = c + 7;
		} else if (strncmp(c, "%version", 8) == 0) {
		    *c = '\0';
		    fprintf(pseudoart, "%s%s", cl, version);
		    cl = c + 8;
		} else if (strncmp(c, "%newsgroup", 10) == 0) {
		    *c = '\0';
		    fprintf(pseudoart, "%s%s", cl, groupname);
		    cl = c + 10;
		} else {
		    *c = '\0';
		    fprintf(pseudoart, "%s%%", cl);
		    cl = c + 1;
		}
	    }
	    fprintf(pseudoart, "%s\n", cl);
	}
	fclose(f);
    } else {
	if (pseudofile)
	    ln_log(LNLOG_SERR, LNLOG_CTOP,
		   "Unable to read pseudoarticle from %s: %m", pseudofile);
	/* fall back to English standard text */
	fprintf(pseudoart,
		"This server is running leafnode, which is a dynamic NNTP proxy.\n"
		"This means that it does not retrieve newsgroups unless someone is\n"
		"actively reading them.\n"
		"\n"
		"If you do an operation on a group - such as reading an article,\n"
		"looking at the group table of contents or similar, then leafnode\n"
		"will go and fetch articles from that group when it next updates.\n"
		"\n"
		"Since you have read this dummy article, leafnode will retrieve\n"
		"the newsgroup %s when fetchnews is run\n"
		"the next time. If you'll look into this group a little later, you\n"
		"will see real articles.\n"
		"\n"
		"If you see articles in groups you do not read, that is almost\n"
		"always because of cross-posting.  These articles do not occupy any\n"
		"more space - they are hard-linked into each newsgroup directory.\n"
		"\n"
		"If you do not understand this, please talk to your newsmaster.\n"
		"\n"
		"Leafnode can be found at http://www.leafnode.org/\n\n",
		groupname);
    }
}

/* build and return an open fd to pseudoart in group */
static FILE *
buildpseudoart(const char *grp)
{
    FILE *f;

    f = tmpfile();
    if (!f) {
	ln_log(LNLOG_SERR, LNLOG_CGROUP,
	       "Could not create pseudoarticle for group %s", grp);
	return f;
    }
    ln_log(LNLOG_SINFO, LNLOG_CGROUP,
	   "Creating pseudoarticle for group %s", grp);
    fprintf(f, "Path: %s!not-for-mail\n", owndn ? owndn : fqdn);
    fprintf(f, "Newsgroups: %s\n", grp);
    fprintf(f, "From: Leafnode <nobody@%s>\n", owndn ? owndn : fqdn);
    fprintf(f, "Subject: Leafnode placeholder for group %s\n", grp);
    fprintf(f, "Date: %s\n", rfctime());
    fprintf(f, "Message-ID: <leafnode:placeholder:%s@%s>\n", grp,
	    owndn ? owndn : fqdn);
    fprintf(f, "\n");
    fprintpseudobody(f, grp);
    rewind(f);
    return f;
}

/* open a pseudo art */
static FILE *
fopenpseudoart(const struct newsgroup *group, const char *arg,
	       const unsigned long article_num)
{
    FILE *f = NULL;
    char msgidbuf[128];
    char *c;
    struct newsgroup *g;

    ln_log(LNLOG_SDEBUG, LNLOG_CGROUP,
	   "fopenpseudoart %s: first %lu last %lu artno %lu",
	   group->name, group->first, group->last, article_num);
    if (article_num && article_num == group->first &&
	group->first >= group->last) {
	f = buildpseudoart(group->name);
    } else if (!article_num) {
	if (!strncmp(arg, "<leafnode:placeholder:", 22)) {
	    strncpy(msgidbuf, arg + 22, 128);
	    msgidbuf[127] = '\0';
	    if ((c = strchr(msgidbuf, '@')) != NULL) {
		*c = '\0';
		g = findgroup(msgidbuf);
		if (g)
		    f = buildpseudoart(g->name);
	    }
	}
    }
    return f;
}

/* return FALSE if group is NOT a pseudo group 
 * return TRUE  if group is a pseudo group 
 */
static int
ispseudogroup(const char *group)
{
    return (!chdirgroup(group, FALSE) && !islocalgroup(group));
}

/* open an article by number or message-id */
static FILE *
fopenart(const struct newsgroup *group, const char *arg, unsigned long *artno)
/* FIXME: when may artno be touched? */
{
    unsigned long int a;
    FILE *f;
    char *t;
    char s[PATH_MAX + 1];	/* FIXME */

    f = NULL;
    t = NULL;
    a = strtoul(arg, &t, 10);
    if (a && t && !*t && group) {
	if (ispseudogroup(group->name)) {
	    f = fopenpseudoart(group, arg, a);
	} else {
	    f = fopen(arg, "r");
	    if (!f && (!islocalgroup(group->name)))
		f = fopenpseudoart(group, arg, a);
	}

	if (f)
	    *artno = a;

	markinterest(group);	/* FIXME: check error */
	/* else try message-id lookup */
    } else if (arg && *arg == '<') {
	f = fopen(lookup(arg), "r");
    } else if (group && *artno) {
	sprintf(s, "%lu", *artno);
	f = fopen(s, "r");
	if (!f)
	    f = fopenpseudoart(group, s, a);
	markinterest(group);	/* FIXME: check error */
    }

    return f;
}

/*
 * Mark an article for download by appending its number to the
 * corresponding file in interesting.groups
 */
static int
markdownload(const struct newsgroup *group, const char *msgid)
{
    char *l;
    FILE *f;
    char s[PATH_MAX + 1];	/* FIXME */

    sprintf(s, "%s/interesting.groups/%s", spooldir, group->name);
    if ((f = fopen(s, "r+"))) {
	while ((l = getaline(f)) != NULL) {
	    if (strncmp(l, msgid, strlen(msgid)) == 0)
		return 0;	/* already marked */
	}
	fprintf(f, "%s\n", msgid);
	ln_log(LNLOG_SDEBUG, LNLOG_CGROUP,
	       "Marking %s: %s for download", group->name, msgid);
	fclose(f);
    }
    return 1;
}

static int
allowposting(void)
{
    static int initialized;
    static char *NOPOSTING;
    static int allowpost = 1;

    /* read NOPOSTING from environment if true */
    if (!initialized) {
	NOPOSTING = getenv("NOPOSTING");
	if (NOPOSTING) {
	    char *p;
	    for (p = NOPOSTING; *p; p++) {
		if (iscntrl((unsigned char) *p)) *p = '_';
	    }
	    allowpost = 0;
	}
    }
    return allowpost;
}

/* display an article or somesuch */
/* DOARTICLE */
static void
doarticle(const struct newsgroup *group, const char *arg, int what,
	  unsigned long *artno)
{
    FILE *f;
    char *p = NULL;
    char *q = NULL;
    unsigned long localartno;
    char *localmsgid = NULL;
    char s[PATH_MAX + 1];	/* FIXME */

    f = fopenart(group, arg, artno);
    if (!f) {
	if (arg && *arg != '<' && !group)
	    nntpprintf("412 No newsgroup selected");
	else if (*arg)
	    nntpprintf("430 No such article: %s", arg);
	else
	    nntpprintf("423 No such article: %lu", *artno);
	return;
    }

    if (!arg) {
	localartno = *artno;
	localmsgid = fgetheader(f, "Message-ID:", 1);
    } else if (*arg == '<') {
	localartno = 0;
	localmsgid = critstrdup(arg, "doarticle");
    } else {
	localartno = strtoul(arg, NULL, 10);
	localmsgid = fgetheader(f, "Message-ID:", 1);
    }

    if (!localartno) {
	/* we have to extract the article number from the Xref: header */
	p = fgetheader(f, "Xref:", 1);
	if (p) {
	    /* search article number of 1st group in Xref: */
	    q = strchr(p, ':');
	    if (q) {
		q++;
		localartno = strtoul(q, NULL, 10);
	    }
	    if (group) {
		/* search article number of group in Xref: */
		q = strstr(p, group->name);
		if (q) {
		    q += strlen(group->name);
		    if (*q++ == ':')
			localartno = strtoul(q, NULL, 10);
		}
	    }
	    free(p);
	}
    }

    sprintf(s, "%3d %lu %s article retrieved - ", 223 - what,
	    localartno, localmsgid);

    if (what == 0)
	strcat(s, "request text separately");
    else if (what == 1)
	strcat(s, "body follows");
    else if (what == 2)
	strcat(s, "head follows");
    else
	strcat(s, "text follows");

    nntpprintf("%s", s);

    /* FIXME: own function */
    while (fgets(s, 1024, f) && *s && (*s != '\n')) {
	if (what & 2) {
	    p = s;
	    if ((p = strchr(p, '\n')))
		*p = '\0';
	    if (*s == '.')
		putc('.', stdout);	/* escape . */
	    fputs(s, stdout);
	    fputs("\r\n", stdout);
	}
    }

    if (what == 3)
	printf("\r\n");		/* empty separator line */

    if (what & 1) {
	if (delaybody && *s != '\n') {
	    if (!markdownload(group, localmsgid)) {
		printf("\r\n\r\n"
		       "\t[Leafnode:]\r\n"
		       "\t[This message has already been "
		       "marked for download.]\r\n");
	    } else {
		printf("\r\n\r\n"
		       "\t[Leafnode:]\r\n"
		       "\t[Message %ld of %s]\r\n"
		       "\t[has been marked for download.]\r\n",
		       localartno, group->name);
	    }
	} else {
	    while (fgets(s, 1024, f) && *s) {
		p = s;
		if ((p = strchr(p, '\n')))
		    *p = '\0';
		printf("%s%s\r\n", *s == '.' ? "." : "", s);
	    }
	}
    }

    if (what)
	printf(".\r\n");

    fclose(f);

    free(localmsgid);
    return;			/* FIXME: OF COURSE there were no errors */
}

/* note bug.. need not be _immediately after_ GROUP */
/* returns 0 for success, errno for error */
static int
markinterest(const struct newsgroup *group)
{
    struct stat st;
    struct utimbuf buf;
    int not_yet_interesting;
    time_t now;
    FILE *f;
    const char *sn[4];
    char *s;

    sn[0] = spooldir;
    sn[1] = "/interesting.groups/";
    sn[2] = group->name;
    sn[3] = 0;			/* do not forget this */

    if (islocalgroup(group->name))
	return 0;		/* local groups don't have to be marked */

    not_yet_interesting = 0;

    s = memstrcat(sn);
    if (s) {
	if (stat(s, &st) == 0) {
	    /* already marked interesting, update atime */
	    now = time(0);
	    buf.actime = (now < st.st_atime) ? st.st_atime : now;
	    buf.modtime = st.st_mtime;
	    /* NOTE: as a side effect, the ctime is also set to now, so effectively, we don't need to care for  */
	    if (utime(s, &buf)) {
		ln_log(LNLOG_SERR, LNLOG_CTOP,
		       "Cannot update timestamp on %s: %m", s);
		not_yet_interesting = 1;
	    }
	} else {
	    not_yet_interesting = 1;
	}

	if (not_yet_interesting) {
	    f = fopen(s, "w");
	    if (f) {
		if (fclose(f)) {
		    int e = errno;
		    ln_log(LNLOG_SERR, LNLOG_CGROUP,
			   "Could not write to %s: %m", s);
		    return e;
		} else {
		    return 0;
		}
	    } else {
		int e = errno;
		ln_log(LNLOG_SERR, LNLOG_CGROUP, "Could not create %s: %m", s);
		return e;
	    }
	}
	free(s);
	return 0;
    } else {
	return errno;
    }
}

/* change to group - note no checks on group name */
static struct newsgroup *
dogroup(const char *arg, unsigned long *artno)
{
    struct newsgroup *g;
    unsigned long i;
    DIR *d;
    struct dirent *de;
    struct stat st;
    int hardlinks;

    rereadactive();
    g = findgroup(arg);
    if (g) {
	if (isinteresting(g->name))
	    markinterest(g);
	if (chdirgroup(g->name, FALSE)) {
	    if (g->count == 0) {
		g->count = (g->last >= g->first ? g->last - g->first + 1 : 0);
		/* FIXME: count articles in group */
		i = 0;
		if (stat(".", &st)) {
		    ln_log(LNLOG_SERR, LNLOG_CTOP,
			   "cannot stat directory of group %s: %m", g->name);
		    hardlinks = 2;
		} else {
		    hardlinks = st.st_nlink;
		}
		if ((d = opendir("."))) {
		    while ((de = readdir(d))) {
			if (de->d_name[0] != '.') {
/* UNIX: skip hidden files */
			    if (!check_allnum(de->d_name)) {
				if (stat(de->d_name, &st)) {
				    ln_log(LNLOG_SWARNING, LNLOG_CGROUP,
					   "warning: cannot stat bogus file "
					   "%s in %s: %m", de->d_name, g->name);
				} else {
				    if (!S_ISDIR(st.st_mode)) {
					if (killbogus) {
					    if (unlink(de->d_name)) {
						ln_log(LNLOG_SERR,
						       LNLOG_CGROUP,
						       "cannot unlink bogus "
						       "file %s in %s: %m",
						       de->d_name, g->name);
					    } else {
						ln_log(LNLOG_SINFO,
						       LNLOG_CGROUP,
						       "unlinked bogus "
						       "file %s in %s",
						       de->d_name, g->name);
					    }
					} else {
					    ln_log(LNLOG_SWARNING,
						   LNLOG_CGROUP,
						   "warning: "
						   "bogus file %s in %s",
						   de->d_name, g->name);
					}

				    } else {
					i++;
				    }
				}
			    } else {
				i++;
			    }
			}
		    }
		    closedir(d);
		    /* UNIX tricks: every subdirectory will have a .., thus,
		       the link count of . is 2 (for . and its name) plus
		       1 for each subdirectory's .. */
		    g->count = i - (hardlinks - 2);
		}
	    }
	    if (g->count == 0) {
		g->first = g->last = 0;	/* FIXME: is this necessary? */
		/* FIXME: is this consistent with other parts (rather
		   set g->first to 1? */
	    }
	} else {		/* group directory is not present */
	    g->first = g->last = g->count = ispseudogroup(g->name) ? 1 : 0;
	}
	freexover();
	xovergroup = 0;
	nntpprintf("211 %lu %lu %lu %s group selected",
		   g->count, g->first, g->last, g->name);
	*artno = g->first;
	fflush(stdout);
	return g;
    } else {
	nntpprintf("411 No such group");
	return 0;
    }
}

static void
dohelp(void)
{
    printf("100 Legal commands\r\n");
/*  printf("  authinfo user Name|pass Password|generic <prog> <args>\r\n"); */
    printf("  authinfo user Name|pass Password\r\n");
    printf("  article [MessageID|Number]\r\n");
    printf("  body [MessageID|Number]\r\n");
/*  printf("  date\r\n"); */
    printf("  group newsgroup\r\n");
    printf("  head [MessageID|Number]\r\n");
    printf("  help\r\n");
/*  printf("  ihave\r\n"); */
    printf("  last\r\n");
/*  printf("  list [active|newsgroups|distributions|schema] [group_pattern]\r\n"); */
    printf("  list [active|newsgroups] [group_pattern]\r\n");
    printf("  list [extensions|overview.fmt]\r\n");
    printf("  listgroup newsgroup\r\n");
    printf("  mode reader\r\n");
    printf("  newgroups yymmdd hhmmss [\"GMT\"] [<distributions>]\r\n");
    printf
	("  newnews newsgroups yymmdd hhmmss [\"GMT\"] [<distributions>]\r\n");
    printf("  next\r\n");
    if (allowposting())
	printf("  post\r\n");
    printf("  quit\r\n");
    printf("  slave\r\n");
    printf("  stat [MessageID|Number]\r\n");
/*  printf("  xgtitle [group_pattern]\r\n"); */
    printf("  xhdr header [range|MessageID]\r\n");
    printf("  xover [range]\r\n");
    printf("  xpat header range|MessageID pat [morepat...]\r\n");
/*  printf("  xpath MessageID\r\n"); */
    printf(".\r\n");
}

static void
domode(const char *arg)
{
    if (!strcasecmp(arg, "reader"))
	nntpprintf("%03d Leafnode %s, pleased to meet you!",
		   201 - allowposting(), version);
    else
	nntpprintf("500 MODE other than READER not supported");
}

static void
domove(const struct newsgroup *group, int by, unsigned long *artno)
{
    char *msgid;
    char s[PATH_MAX + 1];	/* FIXME */

    by = (by < 0) ? -1 : 1;
    if (group) {
	if (*artno) {
	    *artno += by;
	    do {
		sprintf(s, "%lu", *artno);
		msgid = getheader(s, "Message-ID:");
		if (!msgid)
		    *artno += by;
	    } while (!msgid && *artno >= group->first && *artno <= group->last);
	    if (*artno > group->last) {
		*artno = group->last;
		nntpprintf("421 There is no next article");
	    } else if (*artno < group->first) {
		*artno = group->first;
		nntpprintf("422 There is no previous article");
	    } else {
		nntpprintf("223 %lu %s article retrieved", *artno, msgid);
	    }
	    if (msgid)
		free(msgid);
	} else {
	    nntpprintf("420 There is no current article");
	}
    } else {
	nntpprintf("412 No newsgroup selected");
    }
}

/* LIST ACTIVE if what==0, else LIST NEWSGROUPS */
static void
list(struct newsgroup *g, int what, char *pattern)
{
    struct newsgroup *ng;
    char *p;

    /* if pattern exists, convert to lowercase */
    if (pattern) {
	p = pattern;
	while (*p) {
	    *p = tolower((unsigned char)*p);
	    p++;
	}
    }
    ng = g;
    while (ng->name) {
	if (what) {
	    if (!pattern)
		printf("%s\t%s\r\n", ng->name, ng->desc ? ng->desc : "-x-");
	    else if (ngmatch(pattern, ng->name) == 0)
		printf("%s\t%s\r\n", ng->name, ng->desc ? ng->desc : "-x-");
	} else {
	    if (!pattern || (ngmatch(pattern, ng->name) == 0))
		printf("%s %010lu %010lu %c\r\n", ng->name, ng->last,
		       ng->first, ng->status);
	}
	ng++;
    }
}

static void
dolist(char *oarg)
{
    char *p = NULL;
    char *arg = critstrdup(oarg, "dolist");

    if (!strcasecmp(arg, "extensions")) {
	nntpprintf("202 extensions supported follow");
	printf(" OVER\r\n" " PAT\r\n" " LISTGROUP\r\n");
	if (authentication)
	    printf(" AUTHINFO USER\r\n");
	printf(".\r\n");
    } else if (!strcasecmp(arg, "overview.fmt")) {
	nntpprintf("215 information follows");
	printf("Subject:\r\n"
	       "From:\r\n"
	       "Date:\r\n"
	       "Message-ID:\r\n"
	       "References:\r\n"
	       "Bytes:\r\n" "Lines:\r\n" "Xref:full\r\n" ".\r\n");
    } else if (!strcasecmp(arg, "active.times")) {
	nntpprintf("215 Placeholder - Leafnode will fetch groups on demand");
	printf("news.announce.newusers 42 tale@uunet.uu.net\r\n"
	       "news.answers 42 tale@uunet.uu.net\r\n" ".\r\n");
    } else {
	rereadactive();
	if (!active) {
	    ln_log_so(LNLOG_SERR, LNLOG_CTOP,
		      "503 Group information file does not exist!");
	} else if (!*arg || strisprefix(arg, "active")) {
	    nntpprintf("215 Newsgroups in form \"group high low flags\".");
	    if (active) {
		if (!*arg || strlen(arg) == 6)
		    list(active, 0, NULL);
		else {
		    while (*arg && (!isspace((unsigned char)*arg)))
			arg++;
		    while (*arg && isspace((unsigned char)*arg))
			arg++;
		    p = arg;
		    list(active, 0, arg);
		}
	    }
	    printf(".\r\n");
	} else if (strisprefix(arg, "newsgroups")) {
	    nntpprintf("215 Descriptions in form \"group description\".");
	    if (active) {
		if (strlen(arg) == 10)
		    list(active, 1, NULL);
		else {
		    while (*arg && (!isspace((unsigned char)*arg)))
			arg++;
		    while (*arg && isspace((unsigned char)*arg))
			arg++;
		    list(active, 1, arg);
		}
	    }
	    printf(".\r\n");
	} else {
	    nntpprintf("503 Syntax error");
	}
    }
}

static void
donewnews(char *arg)
{
    struct stringlist *l = cmdlinetolist(arg);
    struct stat st;
    struct tm timearray;
    struct tm *ltime;
    time_t age;
    time_t now;
    int year, century;
    int a, b;
    DIR *d, *ng;
    struct dirent *de, *nga;
    char s[PATH_MAX + 1];	/* FIXME */

    if (stringlistlen(l) < 3) {
	nntpprintf("502 Syntax error");
	return;
    }
    now = time(0);
    ltime = localtime(&now);
    year = ltime->tm_year % 100;
    century = ltime->tm_year / 100;	/* 0 for 1900-1999, 1 for 2000-2099 etc */
    memset(&timearray, 0, sizeof(timearray));
    a = strtol(l->next->string, NULL, 10);
    /* NEWNEWS may have the form YYMMDD or YYYYMMDD.
       Distinguish between the two */
    b = a / 10000;
    if (b < 100) {
	/* YYMMDD */
	if (b <= year)
	    timearray.tm_year = b + (century * 100);
	else
	    timearray.tm_year = b + (century - 1) * 100;
    } else if (b < 1000) {
	/* YYYMMDD - happens with buggy newsreaders */
	/* In these readers, YYY=100 is equivalent to YY=00 or YYYY=2000 */
	ln_log(LNLOG_SNOTICE, LNLOG_CSERVER,
	       "NEWGROUPS year is %d: please update your newsreader", b);
	timearray.tm_year = b;
    } else {
	/* YYYYMMDD */
	timearray.tm_year = b - 1900;
    }
    timearray.tm_mon = (a % 10000 / 100) - 1;
    timearray.tm_mday = a % 100;
    a = strtol(l->next->next->string, NULL, 10);
    timearray.tm_hour = a / 10000;
    timearray.tm_min = a % 10000 / 100;
    timearray.tm_sec = a % 100;
    /* mktime() shall guess correct value of tm_isdst(0 or 1) */
    timearray.tm_isdst = -1;
    /* mktime() assumes local time -> correct according to timezone
       (global variable set by last call to localtime()) if GMT is
       not requested */
    age = mktime(&timearray) - (strstr(arg, "GMT") ? gmt_off : 0);
    nntpprintf("230 List of new articles in newsgroup %s", l->string);
    sprintf(s, "%s/interesting.groups", spooldir);
    d = opendir(s);
    if (!d) {
	ln_log(LNLOG_SERR, LNLOG_CTOP, "Unable to open directory %s: %m", s);
	printf(".\r\n");
	return;
    }
    while ((de = readdir(d))) {
	if (ngmatch((const char *)&(l->string), de->d_name) == 0) {
	    chdirgroup(de->d_name, FALSE);
	    getxover();
	    ng = opendir(".");
	    while ((nga = readdir(ng))) {
		long artno;

		if (get_long(nga->d_name, &artno)) {
		    if ((stat(nga->d_name, &st) == 0) &&
			(*nga->d_name != '.') && S_ISREG(st.st_mode) &&
			(st.st_ctime > age)) {
			long xo = findxover(artno);

			if (xo >= 0) {
			    char *x = cuttab(xoverinfo[xo].text, XO_MESSAGEID);
			    if (x) {
				fputs(x, stdout);
				fputs("\r\n", stdout);
				free(x);
			    } else {
				/* FIXME: cannot find message ID in XOVER */
			    }
			} else {
			    /* FIXME: cannot find XOVER for article */
			}
		    }		/* too old */
		}		/* if not a number, not an article */
	    }			/* readdir loop */
	    closedir(ng);
	    free(xoverinfo);
	    xoverinfo = 0;
	}
    }
    closedir(d);
    fputs(".\r\n", stdout);
}

static void
donewgroups(const char *arg)
{
    struct tm timearray;
    struct tm *ltime;
    time_t age;
    time_t now;
    int year, century;
    char *l;
    int a, b;
    struct newsgroup *ng;

    now = time(0);
    ltime = localtime(&now);
    year = ltime->tm_year % 100;
    century = ltime->tm_year / 100;	/* 0 for 1900-1999, 1 for 2000-2099 etc */
    memset(&timearray, 0, sizeof(timearray));
    l = NULL;
    a = strtol(arg, &l, 10);
    /* NEWGROUPS may have the form YYMMDD or YYYYMMDD.
       Distinguish between the two */
    b = a / 10000;
    if (b < 100) {
	/* YYMMDD */
	if (b <= year)
	    timearray.tm_year = b + (century * 100);
	else
	    timearray.tm_year = b + (century - 1) * 100;
    } else if (b < 1000) {
	/* YYYMMDD - happens with buggy newsreaders */
	/* In these readers, YYY=100 is equivalent to YY=00 or YYYY=2000 */
	ln_log(LNLOG_SNOTICE, LNLOG_CGROUP,
	       "NEWGROUPS year is %d: please update your newsreader", b);
	timearray.tm_year = b;
    } else {
	/* YYYYMMDD */
	timearray.tm_year = b - 1900;
    }
    timearray.tm_mon = (a % 10000 / 100) - 1;
    timearray.tm_mday = a % 100;
    while (*l && isspace((unsigned char)*l))
	l++;
    a = strtol(l, NULL, 10);	/* we don't care about the rest of the line */
    timearray.tm_hour = a / 10000;
    timearray.tm_min = a % 10000 / 100;
    timearray.tm_sec = a % 100;
    /* mktime() shall guess correct value of tm_isdst(0 or 1) */
    timearray.tm_isdst = -1;
    /* mktime() assumes local time -> correct according to timezone
       (global variable set by last call to localtime()) if GMT is not
       requested */
    age = mktime(&timearray) - (strstr(arg, "GMT") ? gmt_off : 0);
    nntpprintf("231 List of new newsgroups since %lu follows",
	       (unsigned long)age);
    ng = active;
    while (ng->name) {
	if (ng->age >= age)
	    printf("%s %lu %lu y\r\n", ng->name, ng->first, ng->last);
	ng++;
    }
    printf(".\r\n");
}

/* next bit is copied from INN 1.4 and modified("broken") by agulbra
   mail to Rich $alz <rsalz@uunet.uu.net> bounced */
/* Scale time back a bit, for shorter Message-ID's. */
#define OFFSET	673416000L
static char ALPHABET[] = "0123456789abcdefghijklmnopqrstuv";
static char *
generateMessageID(void)
{
    static char buff[80];
    static time_t then;
    static unsigned int fudge;
    time_t now;
    char *p;
    int n;

    now = time(0);		/* might be 0, in which case fudge 
				   will almost fix it */
    if (now != then)
	fudge = 0;
    else
	fudge++;
    then = now;
    p = buff;
    *p++ = '<';
    n = (unsigned int)now - OFFSET;
    while (n) {
	*p++ = ALPHABET[(int)(n & 31)];
	n >>= 5;
    }
    *p++ = '.';
    n = fudge * 32768 + (int)getpid();
    while (n) {
	*p++ = ALPHABET[(int)(n & 31)];
	n >>= 5;
    }
    sprintf(p, ".ln@%s>", owndn ? owndn : fqdn);
    return (critstrdup(buff, "generateMessageID"));
}

/* the end of what I stole from rsalz and then mangled */

/* checks for malformatted newsgroups header */
/* returns 0 if malformatted (spurious whitespace) */
static int
validate_newsgroups(const char *n)
{
    char *t;

    if (strisprefix(n, "Newsgroups:")) {
	n += 11;
	/* skip separating whitespace */
	n += strspn(n, WHITESPACE);
    }

    /* find other whitespace */
    t = strpbrk(n, WHITESPACE);
    if (t) {
#if 0
	/* tolerant version */
	/* we have found whitespace, check if it's just trailing */
	if (strspn(t, WHITESPACE) != strlen(t, WHITESPACE))
	    return 0;
#else
	/* pedantic version */
	return 0;
#endif
    }
    return 1;
}

static void
dopost(void)
{
    char *line, *msgid;
    int havefrom = 0;
    int havepath = 0;
    int havedate = 0;
    int havebody = 0;
    int havenewsgroups = 0;
    int havemessageid = 0;
    int ngmalformatted = 0;
    int havesubject = 0;
    int err = 0;
    int duplicate = FALSE;
    int hdrtoolong = FALSE;
    size_t i, len;
    int out;
    char inname[PATH_MAX + 1];
    static int postingno;	/* starts at 0 */

    do {
	sprintf(inname, "%s/in.coming/%d-%lu-%d",
		spooldir, (int)getpid(), (unsigned long)time(NULL),
		++postingno);
	out = open(inname, O_WRONLY | O_EXCL | O_CREAT, (mode_t) 0444);
    } while (out < 0 && errno == EEXIST);

    if (out < 0) {
	ln_log_so(LNLOG_SERR, LNLOG_CTOP,
		  "441 Unable to open spool file %s: %m", inname);
	return;
    }

    msgid = generateMessageID();
    nntpprintf("340 Ok, recommended ID %s", msgid);
    fflush(stdout);
    /* get headers */
    do {
	line = getaline(stdin);

	if (!line) {
	    /* client died */
	    log_unlink(inname);
	    nntpprintf("400 Premature end of input. Exiting.");
	    exit(0);
	}

	/* premature end (no body) */
	if (!strcmp(line, ".")) {
	    err = TRUE;
	    break;
	}

	if (strisprefix(line, "From:")) {
	    if (havefrom)
		err = TRUE;
	    else
		havefrom = TRUE;
	}

	if (strisprefix(line, "Path:")) {
	    if (havepath)
		err = TRUE;
	    else
		havepath = TRUE;
	}

	if (strisprefix(line, "Message-ID:")) {
	    if (havemessageid)
		err = TRUE;
	    else {
		struct stat st;

		havemessageid = TRUE;
		free(msgid);
		msgid = mgetheader("Message-ID:", line);
		if (stat(lookup(msgid), &st) == 0) {
		    err = TRUE;
		    duplicate = TRUE;
		}
	    }
	}

	if (strisprefix(line, "Subject:")) {
	    if (havesubject)
		err = TRUE;
	    else
		havesubject = TRUE;
	}

	if (strisprefix(line, "Newsgroups:")) {
	    if (havenewsgroups)
		err = TRUE;
	    else {
		if (validate_newsgroups(line))
		    havenewsgroups = TRUE;
		else
		    err = ngmalformatted = TRUE;
	    }
	}

	if (strisprefix(line, "Date:")) {
	    if (havedate)
		err = TRUE;
	    else
		havedate = TRUE;
	}

	len = strlen(line);
	/* maximal header length of 1000 chars is suggested in RFC.
	   Some people have requested that leafnode should barf if
	   headers are too long. If you don't want this, just remove
	   the following lines.
	 */

	if (len > 1000) {
	    hdrtoolong = TRUE;
	    err = TRUE;
	}
	/* replace tabs with spaces */
	for (i = 0; i < len; i++)
	    if (isspace((unsigned char)line[i]))
		line[i] = ' ';
	if (len && line[len - 1] == '\n')
	    line[--len] = '\0';
	if (len && line[len - 1] == '\r')
	    line[--len] = '\0';
	if (len && line[len - 1] == '\r')
	    line[--len] = '\0';
	if (len)
	    write(out, line, len);
	else {
	    if (!havepath) {
		writes(out, "Path: ");
		if (owndn)
		    writes(out, owndn);
		else
		    writes(out, fqdn);
		writes(out, "!not-for-mail\r\n");
	    }
	    if (!havedate) {
		const char *l = rfctime();

		writes(out, "Date: ");
		writes(out, l);
		writes(out, "\r\n");
	    }
	    if (!havemessageid) {
		char tmp[800];

		snprintf(tmp, sizeof(tmp), "Message-ID: %s\r\n", msgid);
		writes(out, tmp);
	    }
	}
	writes(out, "\r\n");
    } while (*line);

    free(msgid);
    /* get bodies */
    if (strcmp(line, ".")) {	/* skip if header contained a single dot line */
	havebody = TRUE;
	do {
	    debug = 0;
	    line = getaline(stdin);
	    debug = debugmode;
	    if (!line) {
		log_unlink(inname);
		exit(EXIT_FAILURE);
	    }
	    len = strlen(line);
	    if (len && line[len - 1] == '\n')
		line[--len] = '\0';
	    if (len && line[len - 1] == '\r')
		line[--len] = '\0';
	    if (len && line[len - 1] == '\r')
		line[--len] = '\0';
	    if (line[0] == '.') {
		/* escape or skip if single dot */
		if (len > 1) {
		    write(out, line + 1, len - 1);
		    writes(out, "\r\n");
		}
	    } else {
		write(out, line, len);
		writes(out, "\r\n");
	    }
	} while (strcmp(line, "."));
	/* safely write to disk before reporting success */
	if (log_fsync(out))
	    err = 1;
	if (log_close(out))
	    err = 1;
    }

    if (havefrom && havesubject && havenewsgroups && !err) {
	FILE *f;
	char *mid = 0;
	char *groups = 0;
	char *outbasename;
	char *forbidden = 0;
	char *modgroup = 0;
	char *moderator = 0;
	char *approved = 0;

	f = fopen(inname, "r");
	if (f) {
	    mid = fgetheader(f, "Message-ID:", 1);
	    groups = fgetheader(f, "Newsgroups:", 1);
	    log_fclose(f);
	}

	if (!f || !mid || !groups) {
	    /* heck, cannot read the file just written */
	    if (!f)
		ln_log(LNLOG_SERR, LNLOG_CTOP,
		       "cannot open %s for reading: %m", inname);
	    else
		ln_log(LNLOG_SERR, LNLOG_CTOP,
		       "cannot read message-id or newsgroups from %s", inname);
	    nntpprintf("503 Could not reopen article, see syslog");
	    log_unlink(inname);
	    return;
	}

	if ((forbidden = checkstatus(groups, 'n'))) {
	    /* Newsgroups: contains a group to which posting is not allowed */
	    ln_log(LNLOG_SNOTICE, LNLOG_CARTICLE,
		   "Article was posted to non-writable group %s", forbidden);
	    nntpprintf("441 Posting to %s not allowed", forbidden);
	    free(forbidden);
	    log_unlink(inname);
	    free(mid);
	    free(groups);
	    return;
	}

	if ((modgroup = checkstatus(groups, 'm'))) {
	    /* Post to a moderated group */
	    moderator = getmoderator(modgroup);
	    if (!moderator && islocal(modgroup)) {
		ln_log(LNLOG_SERR, LNLOG_CTOP,
		       "Did not find moderator address for local moderated "
		       "group %s", modgroup);
		nntpprintf("503 Configuration error: No moderator for %s",
			   modgroup);
		free(mid);
		free(groups);
		free(modgroup);
		log_unlink(inname);
		return;
	    }
	    approved = getheader(inname, "Approved:");
	}

	if (!moderator && !islocal(groups)) {
	    /* also posted to external groups or moderated group with
	       unknown moderator, store into out.going */
	    char s[PATH_MAX + 1];	/* FIXME: overflow possible */
	    outbasename = strrchr(inname, '/');
	    outbasename++;
	    sprintf(s, "%s/out.going/%s", spooldir, outbasename);
	    if (sync_link(inname, s)) {
		ln_log(LNLOG_SERR, LNLOG_CARTICLE,
		       "unable to schedule for posting to upstream, "
		       "link %s -> %s failed: %m", inname, s);
		nntpprintf("503 Could not schedule article for posting "
			   "to upstream, see syslog.");
		log_unlink(inname);
		return;
	    }
	    if (modgroup && !approved) {
		free(modgroup);
		log_unlink(inname);
		return;
	    }
	}

	if (moderator && !approved) {
	    /* Mail the article to the moderator */
	    int fd;

	    fd = open(inname, O_RDONLY);
	    if (fd < 0) {
		ln_log(LNLOG_SERR, LNLOG_CTOP, "Couldn't open article %s: %m",
		       inname);
	    } else if (mailto(moderator, fd)) {
		nntpprintf("503 Posting to moderator %s failed: %m", moderator);
	    } else {
		ln_log(LNLOG_SDEBUG, LNLOG_CARTICLE,
		       "Message %s mailed to moderator %s", inname, moderator);
		nntpprintf("240 Article mailed to moderator");
	    }

	    if (fd >= 0)
		log_close(fd);
	    free(moderator);
	    free(mid);
	    free(groups);
	    log_unlink(inname);
	    return;
	}
	if (approved)
	    free(approved);

	ln_log(LNLOG_SINFO, LNLOG_CTOP, "%s POST %s %s",
	       islocal(groups) ? "LOCAL" : "UPSTREAM", mid, groups);

	switch (fork()) {
	case -1:
	    /* parent (fork failed) */
	    ln_log(LNLOG_SERR, LNLOG_CTOP, "dopost: fork: %m, "
		   "local post will show up delayed.");
	    nntpprintf("240 Posting accepted, but local post will show up "
		       "delayed, see syslog.");
	    break;

	case 0:
	    /* child */
	    fclose(stdin);
	    fclose(stdout);
	    fclose(stderr);
	    if (lockfile_exists(FALSE)) {
		ln_log(LNLOG_SNOTICE, LNLOG_CARTICLE,
		       "Cannot obtain lock file to store article %s. "
		       "It will be posted later.", inname);
		_exit(EXIT_FAILURE);
	    }

	    rereadactive();
	    feedincoming();
	    writeactive();
	    log_unlink(lockfile);
	    _exit(0);

	default:
	    nntpprintf("240 Article posted, now be patient");
	    break;
	}			/* switch(fork()) */
	free(mid);
	free(groups);
	return;
    }

    log_unlink(inname);
    if (!havefrom)
	nntpprintf("441 From: header missing, article not posted");
    else if (!havesubject)
	nntpprintf("441 Subject: header missing, article not posted");
    else if (ngmalformatted)
	nntpprintf
	    ("441 Spurious whitespace in Newsgroups: header, fix your news reader");
    else if (!havenewsgroups)
	nntpprintf("441 Newsgroups: header missing, article not posted");
    else if (hdrtoolong)
	nntpprintf("441 Header too long, article not posted");
    else if (duplicate)
	nntpprintf("441 435 Duplicate, article not posted");
    else if (!havebody)
	nntpprintf("441 Article has no body, not posted");
    else
	nntpprintf("441 Error, article not posted, see syslog.");
}

void
doxhdr(const struct newsgroup *group, const char *arg, unsigned long artno)
{
    /* NOTE: XPAT is not to change the current article pointer, thus,
       we're using call by value here */
    struct stringlist *l = cmdlinetolist(arg);

    switch (stringlistlen(l)) {
    case 1:
	doselectedheader(group, l->string, NULL, NULL, &artno);
	/* discard changes to artno */
	break;
    case 2:
	doselectedheader(group, l->string, l->next->string, NULL, &artno);
	/* discard changes to artno */
	break;
    default:
	nntpprintf("502 Usage: XHDR header [{first[-[last]]|<message-id>}]");
    }
    if (l)
	freelist(l);
}

void
doxpat(const struct newsgroup *group, const char *arg, unsigned long artno)
{
    /* NOTE: XPAT is not to change the current article pointer, thus,
       we're using call by value here */
    struct stringlist *l = cmdlinetolist(arg);

    if (stringlistlen(l) < 3) {
	nntpprintf("502 Usage: XPAT header first[-[last]] pattern or "
		   "XPAT header message-id pattern");
    } else {
	doselectedheader(group, l->string, l->next->string, l->next->next,
			 &artno);
	/* discard changes to artno */
    }
    if (l)
	freelist(l);
}

/**
 * This function outputs a list of article headers.
 * You can give a list of patterns, then only headers matching one of
 * these patterns are listed.  If you give NULL as the patterns
 * argument, every header is listed.
 */
void
doselectedheader(const struct newsgroup *group /** current newsgroup */ ,
		 const char *hd /** header to extract */ ,
		 const char *messages /** message range */ ,
		 struct stringlist *patterns /** pattern */ ,
		 unsigned long *artno /** currently selected article */ )
{
    /* FIXME: this is bloody complex and hard to follow */
    const char *const h[] = { "Subject:", "From:", "Date:", "Message-ID:",
	"References:", "Bytes:", "Lines:"
    };
    int OVfield;
    char *l;
    unsigned long a, b = 0, c;
    long int i;
    FILE *f;
    struct stringlist *ap;
    char *header;

    header = (char *)critmalloc((i = strlen(hd)) + 2, "doselectedheader");
    strcpy(header, hd);
    if (header[i - 1] != ':')
	strcpy(header + i, ":");

    /* HANDLE MESSAGE-ID FORM */
    if (messages && *messages == '<') {
	f = fopenart(group, messages, artno);
	if (!f) {
	    nntpprintf("430 No such article");
	    return;
	}
	l = fgetheader(f, header, 1);
	if (!l || !(*l)) {
	    nntpprintf("221 No such header: %s", hd);
	    printf(".\r\n");
	    fclose(f);
	    return;
	}
	STRIP_TRAILING_SPACE(l);
	if (patterns) {
	    ap = patterns;
	    while (ap) {
		if (0 == ngmatch((const char *)&(ap->string), l))
		    break;
		ap = ap->next;
	    }
	    if (!ap) {		/* doesn't match any pattern */
		nntpprintf("221 %s matches follow:", hd);
		printf(".\r\n");
		fclose(f);
		return;
	    }
	}
	nntpprintf("221 %s %s follow:", hd, (patterns ? "matches" : "headers"));
	printf("%s %s\r\n.\r\n", messages, l ? l : "");
	fclose(f);
	return;
    }

    /* normal range form chosen */

    if (!group) {
	nntpprintf("412 Use the GROUP command first");
	return;
    }

    markinterest(group);

    /* check if header can be obtained from overview */
    OVfield = sizeof(h) / sizeof(h[0]);
    do {
	OVfield--;
    } while (OVfield >= 0 && strcasecmp(h[OVfield], header));

    if (!ispseudogroup(group->name)) {
	/* FIXME: does this work for local groups? */
	/* is a real group */
	if (xovergroup != group) {
	    if (getxover())
		xovergroup = group;
	}
    } else {
	/* is a pseudo group */

	if (patterns) {		/* placeholder matches pseudogroup never */
	    nntpprintf("221 %s header matches follow:", hd);
	    printf(".\r\n");
	    return;
	}

	if (OVfield < 0) {
	    if (!strcasecmp(header, "Newsgroups:")) {
		nntpprintf("221 First line of %s pseudo-header follows:", hd);
		printf("1 %s\r\n", group->name);
		printf(".\r\n");
	    } else if (!strcasecmp(header, "Path:")) { 
		nntpprintf("221 First line of %s pseudo-header follows:", hd);
		printf("1 %s!not-for-mail\r\n", owndn ? owndn : fqdn);
		printf(".\r\n");
	    } else {
		nntpprintf("221 No such header: %s", hd);
		printf(".\r\n");
	    }
	    return;
	}

	nntpprintf("221 First line of %s pseudo-header follows:", hd);
	if (OVfield == 0)	/* Subject: */
	    printf("1 Leafnode placeholder for group %s\r\n", group->name);
	else if (OVfield == 1)	/* From: */
	    printf("1 Leafnode <nobody@%s>\r\n", owndn ? owndn : fqdn);
	else if (OVfield == 2)	/* Date: */
	    printf("1 %s\r\n", rfctime());
	else if (OVfield == 3)	/* Message-ID: */
	    printf("1 <leafnode:placeholder:%s@%s>\r\n", group->name,
		   owndn ? owndn : fqdn);
	else if (OVfield == 4)	/* References */
	    printf("1 (none)\r\n");	/* FIXME */
	else if (OVfield == 5)	/* Bytes */
	    printf("1 %d\r\n", 1024);	/* just a guess */
	else if (OVfield == 6)	/* Lines */
	    printf("1 %d\r\n", 22);	/* FIXME: from buildpseudoart() */
	printf(".\r\n");
	return;
    }

    /* HANDLE ARTNO/RANGE FORM */
    if (messages) {
	if (parserange(messages, &a, &b) & RANGE_ERR) {
	    nntpprintf("502 Syntax error");
	    return;
	}

	if (!dorange(messages, &a, &b, *artno, group->first, group->last))
	    return;
    } else {
	a = b = *artno;
    }

    /* Check whether there are any articles in the range specified */
    i = -1;
    c = a;
    while ((c <= b) && (i == -1)) {
	i = findxover(c);
	c++;
    }
    if (i == -1) {
	nntpprintf("420 No article in specified range.");
	return;
    }
    if (OVfield >= 0) {
	nntpprintf("221 %s header %s(from overview) for postings %lu-%lu:",
		   hd, patterns ? "matches " : "", a, b);

	for (c = a; c <= b; c++) {
	    char *t = 0;
	    if (xoverinfo &&
		c >= xfirst && c <= xlast && (i = findxover(c)) >= 0) {
		char *li = xoverinfo[i].text;
		int d;

		/* OK MA 2001-06-27 */
		for (d = 0; li && d <= OVfield; d++) {
		    li = strchr(li, '\t');
		    if (li)
			li++;
		}

		if (li) {
		    char *p;
		    p = strchr(li, '\t');
		    if (!p)
			p = li + strlen(li);

		    t = (char *)critmalloc(p - li + 1, "doselectedheader");
		    mastrncpy(t, li, p - li + 1);
		}

		if (patterns) {
		    ap = patterns;
		    while (ap) {
			if (!ngmatch((const char *)&(ap->string), t))
			    break;
			ap = ap->next;
		    }
		    if (!ap) {
			free(t);
			continue;
		    }
		}

		if (t) {
		    printf("%lu %s\r\n", c, t);
		    free(t);
		}
	    }
	}
    } else {
	nntpprintf
	    ("221 %s header %s(from article files) for postings %lu-%lu:",
	     hd, patterns ? "matches " : "", a, b);
	for (c = a; c <= b; c++) {
	    char s[64];

	    sprintf(s, "%lu", c);
	    l = getheader(s, header);
	    if (l) {
		STRIP_TRAILING_SPACE(l);
		if (patterns) {
		    ap = patterns;
		    while (ap) {
			if (!ngmatch((const char *)&(ap->string), l))
			    break;
			ap = ap->next;
		    }
		    if (!ap)
			continue;
		}
		if (*l)
		    printf("%lu %s\r\n", c, l);
		free(l);
	    }
	}
    }
    printf(".\r\n");
    return;
}

/** call this only if a previous parserange with the same arg yielded no
 * error return! 
 *
 * \return 
 *  - 1 for success 
 *  - 0 if no articles are found (this is printed)
 */
static int
dorange(const char *arg,
	unsigned long *a, unsigned long *b,
	unsigned long artno, unsigned long lo, unsigned long hi)
{
    int i;

    if (arg) {
	*a = lo;
	*b = hi;
	i = parserange(arg, a, b);
	if (!(i & RANGE_HAVETO))
	    *b = *a;
    } else {
	*a = *b = artno;
    }

    if ((*b < *a) || (*a > hi) || (*b < lo)) {
	nntpprintf("420 No articles in specified range.");
	return 0;
    }

    /* sanitize */
    if (*b > hi)
	*b = hi;
    if (*a < lo)
	*a = lo;

    return 1;
}

/** implement XOVER command. */
void
doxover(const struct newsgroup *group, const char *arg, unsigned long artno)
{
    char *l;
    unsigned long a, b, art;
    long int idx;
    int flag = FALSE;
    int i;

    if (!group) {
	nntpprintf("412 Use the GROUP command first");
	return;
    }

    markinterest(group);
    l = NULL;

    if (!*arg)
	arg = 0;		/* simplify subsequent tests */

    if (arg) {
	i = parserange(arg, &a, &b);
	if (i & RANGE_ERR) {
	    nntpprintf("502 Usage: XOVER [first[-[last]]]");
	    return;
	}
    }

    if (!ispseudogroup(group->name)) {
	if (xovergroup != group) {
	    freexover();
	    if (getxover())
		xovergroup = group;
	    else
		xovergroup = 0;
	}

	if (!dorange(arg, &a, &b, artno, xfirst, xlast))
	    return;

	for (art = a; art <= b; art++) {
	    idx = findxover(art);
	    if (idx >= 0 && xoverinfo[idx].text != NULL) {
		if (!flag) {
		    flag = TRUE;
		    nntpprintf
			("224 Overview information for postings %lu-%lu:",
			 a, b);
		}
		printf("%s\r\n", xoverinfo[idx].text);
	    }
	}
	if (flag)
	    printf(".\r\n");
	else
	    nntpprintf("420 No articles in specified range.");
    } else {
	/* _is_ pseudogroup */
	if ((b < 1) || (a > 1) || (a > b)) {
	    nntpprintf("420 No articles in specified range.");
	    return;
	}
	nntpprintf("224 Overview information (pseudo) for postings 1-1:");
	nntpprintf("%lu\t"
		   "Leafnode placeholder for group %s\t"
		   "nobody@%s (Leafnode)\t%s\t"
		   "<leafnode.%s@%s>\t\t1000\t40",
		   b, group->name,
		   owndn ? owndn : fqdn, rfctime(), group->name,
		   owndn ? owndn : fqdn);
	printf(".\r\n");
    }
}

struct newsgroup *
dolistgroup(struct newsgroup *group, const char *arg, unsigned long *artno)
{
    struct newsgroup *g;
    int emptygroup = FALSE;
    long int idx;
    unsigned long art;
    int pseudogroup;

    if (arg && strlen(arg)) {
	g = findgroup(arg);
	if (!g) {
	    nntpprintf("411 No such group: %s", arg);
	    return 0;
	} else {
	    group = g;
	    *artno = g->first;
	}
    } else if (group) {
	g = group;
    } else {
	nntpprintf("412 No group specified");
	return 0;
    }
    group = g;
    if ((pseudogroup = ispseudogroup(g->name))) {
	/* group has not been visited before */
	markinterest(group);
    } else if ((xovergroup != group) && !getxover()) {
	if (isinteresting(g->name)) {
	    /* group has already been marked as interesting but is empty */
	    emptygroup = TRUE;
	}
    } else {
	xovergroup = group;
    }
    markinterest(group);
    if (pseudogroup) {
	nntpprintf("211 Article list for %s follows (pseudo)", g->name);
	printf("%lu \r\n", g->last ? g->last : 1);
    } else if (emptygroup) {
	nntpprintf("211 No articles in %s", g->name);
    } else {
	nntpprintf("211 Article list for %s (%lu-%lu) follows",
		   g->name, xfirst, xlast);
	for (art = xfirst; art <= xlast; art++) {
	    idx = findxover(art);
	    if (idx >= 0 && xoverinfo[idx].text)
		printf("%lu \r\n", art);
	}
    }
    fputs(".\r\n", stdout);
    return group;
}

/*
 * authinfo stuff
 */
/*
 * read users and their crypted passwords into a stringlist
 */
static int
readpasswd(void)
{
    char *l;
    FILE *f;
    int error;
    struct stringlist *ptr = NULL;
    char s[PATH_MAX + 1];	/* FIXME */

    sprintf(s, "%s/users", libdir);
    if ((f = fopen(s, "r")) == NULL) {
	error = errno;
	ln_log(LNLOG_SERR, LNLOG_CTOP, "unable to open %s: %m", s);
	authentication = AM_NAME;
	return error;
    }
    while ((l = getaline(f)) != NULL) {
	appendtolist(&users, &ptr, l);
    }
    fclose(f);
    return 0;
}

static int
isauthorized(void)
{
    if (!authentication)
	return TRUE;
    if (authflag)
	return TRUE;
    nntpprintf("480 Authentication required for command");
    return FALSE;
}

/** Bug: this must be replaced by a proper state machine */
void
doauthinfo(const char *arg)
{
    char cmd[MAXLINELENGTH] = "";
    char param[MAXLINELENGTH] = "";
    static char *user = NULL;
    char *p;
    int result = 0;

    if (!arg || !strlen(arg))
	result = P_SYNTAX_ERROR;
    else if (!authentication)
	result = P_NOT_SUPPORTED;
    else if (sscanf(arg, "%s %s", cmd, param) != 2)
	result = P_SYNTAX_ERROR;
    else if (!strcasecmp(cmd, "user")) {
	if (authentication == AM_NAME) {
	    if (getpwnam(param))
		result = P_ACCEPTED;
	    else
		result = P_REJECTED;
	} else if (authentication == AM_FILE) {
	    user = critstrdup(param, "doauthinfo");
	    result = P_CONTINUE;
	}
    } else if (!strcasecmp(cmd, "pass")) {
	if (authentication == AM_NAME) {
	    result = P_SYNTAX_ERROR;
	} else if (authentication == AM_FILE) {
	    if ((p = findinlist(users, user)) != NULL) {
		char templ[80];
		char passw[80];

		snprintf(templ, 79, "%s %%s", user);
		if (sscanf(p, templ, passw) != 1) {
		    result = P_SYNTAX_ERROR;
		    ln_log(LNLOG_SWARNING, LNLOG_CTOP,
			   "Could not parse line %s", p);
		} else {
		    p = crypt(param, passw);
		    if (!strcmp(passw, p))
			result = P_ACCEPTED;
		    else {
			result = P_REJECTED;
			printf("%s\n", p);
		    }
		}
	    } else {
		/* user unknown */
		result = P_REJECTED;
	    }
	} else {
	    ln_log(LNLOG_SERR, LNLOG_CTOP,
		   "This should not happen: doauthinfo " __FILE__
		   " line %d", __LINE__);
	}
    } else {
	result = P_SYNTAX_ERROR;
    }
    switch (result) {
    case P_ACCEPTED:
	nntpprintf("%d Authentication accepted", result);
	authflag = 1;
	break;
    case P_CONTINUE:
	nntpprintf("%d More authentication required", result);
	break;
    case P_REJECTED:
	nntpprintf("%d Authentication rejected", result);
	authflag = 0;
	break;
    case P_SYNTAX_ERROR:
	nntpprintf("%d Authentication method not supported", result);
	break;
    case P_NOT_SUPPORTED:
	nntpprintf("%d Authentication not supported", result);
	break;
    default:
	nntpprintf("%d This should not happen: %d", P_SYNTAX_ERROR, result);
	break;
    }				/* switch */
}

static void
log_sockaddr(const char *tag, const struct sockaddr *sa)
{
    int h_e;
    char *s = masock_sa2name(sa, &h_e);
    char *a = masock_sa2addr(sa);

    if (!a) {
	ln_log(LNLOG_SERR, LNLOG_CTOP, "cannot convert address to string");
	exit(EXIT_FAILURE);
    }
    if (!s)
	ln_log(LNLOG_SNOTICE, LNLOG_CTOP,
	       "cannot get hostname: %s (h_errno=%d)", my_h_strerror(h_e), h_e);
    ln_log(LNLOG_SINFO, LNLOG_CTOP, "%s: %s (%s)", tag, s ? s : a, a);
    if (s)
	free(s);
    free(a);			/* a == NULL caught above */
}

/* this dummy function is used so we can define a no-op for SIGCHLD */
static void dummy(int unused);
static void
dummy(int unused)
{
    (void)unused;
}

int
main(int argc, char **argv)
{
    int option, reply, err;
    socklen_t fodder;
    char conffile[PATH_MAX + 1];
    FILE *se;
    const long bufsize = BLOCKSIZE;
    char *buf = (char *)critmalloc(bufsize, "main_loop");

#ifdef HAVE_IPV6
    struct sockaddr_in6 sa, peer;
#else
    struct sockaddr_in sa, peer;
#endif

    whoami();

    /* set buffer */
    fflush(stdout);
    setvbuf(stdout, buf, _IOFBF, bufsize);

    if (((err = snprintf(conffile, sizeof(conffile), "%s/config", libdir)) < 0)
	|| (err >= (int)sizeof(conffile))) {
	exit(EXIT_FAILURE);
    }

    if (!initvars(argv[0]))
	exit(EXIT_FAILURE);
    ln_log_open("leafnode");

    /* have stderr discarded */
    se = freopen("/dev/null", "w+", stderr);
    if (!se) {
	ln_log_so(LNLOG_SERR, LNLOG_CTOP,
		  "503 Failure: cannot open /dev/null: %m");
	exit(EXIT_FAILURE);
    }

    while ((option = getopt(argc, argv, "F:DV")) != -1) {
	if (!parseopt("leafnode", option, optarg, conffile, sizeof(conffile))) {
	    ln_log(LNLOG_SWARNING, LNLOG_CTOP, "Unknown option %c", option);
	    exit(EXIT_FAILURE);
	}
	debug = debugmode;
	verbose = 0;		/* overwrite verbose logging */
    }

    if ((reply = readconfig(conffile)) != 0) {
	ln_log_so(LNLOG_SERR, LNLOG_CTOP,
		  "503 Unable to read configuration from %s: %s",
		  conffile, strerror(reply));
	exit(EXIT_FAILURE);
    }

    verbose = 0;
    umask((mode_t) 02);

    /* get local address */
    fodder = sizeof(sa);
    if (getsockname(0, (struct sockaddr *)&sa, (socklen_t *) & fodder)) {
	if (errno != ENOTSOCK) {
	    ln_log(LNLOG_SNOTICE, LNLOG_CTOP, "cannot getsockname: %m");
	    printf("503 Cannot getsockname (%s), aborting\r\n",
		   strerror(errno));
	    exit(EXIT_FAILURE);
	}
    } else {
	log_sockaddr("local  host", (struct sockaddr *)&sa);
    }

    /* get remote address */

    fodder = sizeof(peer);
    if (getpeername(0, (struct sockaddr *)&peer, (socklen_t *) & fodder)) {
	if (errno != ENOTSOCK) {
	    ln_log(LNLOG_SERR, LNLOG_CTOP, "Connect from unknown client: %m");
	    printf("503 Cannot getpeername (%s), aborting\r\n",
		   strerror(errno));
	    exit(EXIT_FAILURE);
	}
    } else {
	log_sockaddr("remote host", (struct sockaddr *)&peer);
    }

/* #endif */
    if (authentication && (reply = readpasswd()) > 0) {
	ln_log_so(LNLOG_SNOTICE, LNLOG_CTOP,
		  "503 Exiting, unable to read user list: %m");
	exit(EXIT_FAILURE);
    }
    rereadactive();
    signal(SIGCHLD, dummy);	/* SIG_IGN would cause ECHILD on wait, so we use
				   our own no-op dummy */
    gmt_off = gmtoff();		/* get difference between local time and GMT */
    printf("%03d Leafnode NNTP daemon, version %s at %s%s %s\r\n",
	   201 - allowposting(), version, owndn ? owndn : fqdn,
	   allowposting() ? "" : " (No posting.)",
	   allowposting() ? "" : getenv("NOPOSTING"));
    fflush(stdout);
    main_loop();
    freexover();
    freeactive();
    freelocal();
    freeconfig();
    free(buf);
    exit(0);
}
